from typing import Any, Type, Union

from pymontrace.tracee import Quantization

def create(fd: int) -> Any: ...
def create_agg_buffer(fd: int, name: str) -> Any: ...
def open_agg_buffer(fd: int) -> Any: ...
def encode_entry(key_data: bytes, value, Quantization: Type) -> bytes: ...
def encode_value(value, Quantization: Type) -> bytes: ...
def decode_value(data, Quantization: Type) -> Union[int, float, Quantization]: ...
